# Copyright 2024 CVS Health and/or one of its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from langfair.metrics.recommendation.metrics.baseclass.metrics import Metric


class JaccardSimilarity(Metric):
    def __init__(self) -> None:
        """
        This class computes the Jaccard Similarity score.
        """
        self.name = "Jaccard"

    def evaluate(self, list1: List[str], list2: List[str]) -> float:
        """
        This method computes Jaccard Similarity between two lists.

        Parameters
        ----------
        list1 : list of strings
            A list of recommendation from an LLM model.

        list2 : list of strings
            Another list of recommendation from an LLM model.

        Returns
        -------
        float
            Jaccard similarity for the two provided lists of recommendations (float)
        """
        x = set(list1)
        y = set(list2)
        return len(x & y) / len(x | y)



# Copyright 2024 CVS Health and/or one of its affiliates
#
# Copyright OpenAI 2015-2024
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# The original work of OpenAI has been modified by CVS Health
# to include costs for only a subset of OpenAI models.

"""Storage place for constants related to cost information such as LLM invokation token cost."""

################################################################################
# A dictionary to store the cost information for different OpenAI models. It maps the model names to their respective input and output token costs.
################################################################################
COST_MAPPING = {
    "gpt-3.5-turbo-0613": {"input": 0.0000015, "output": 0.000002},
    "gpt-3.5-turbo-0301": {"input": 0.000002, "output": 0.000002},
    "gpt-3.5-turbo-16k-0613": {"input": 0.000003, "output": 0.000004},
    "gpt-3.5-turbo-16k-1106": {"input": 0.000001, "output": 0.000002},
    "gpt-3.5-turbo-0125": {"input": 0.0000005, "output": 0.0000015},
    "gpt-4-0314": {"input": 0.00003, "output": 0.00006},
    "gpt-4-0613": {"input": 0.00003, "output": 0.00006},
    "gpt-4-32k-0314": {"input": 0.00006, "output": 0.00012},
    "gpt-4-32k-0613": {"input": 0.00006, "output": 0.00012},
    "gpt-4-turbo": {"input": 0.00001, "output": 0.00003},
    "gpt-4-turbo-2024-04-09": {"input": 0.00001, "output": 0.00003},
}
FAILURE_MESSAGE = "Unable to get response"
TOKEN_COST_DATE = "10/21/2024"  # nosec - for some reason bandit thinks this is a hardcoded secret #noqa - bandit: B105



# Copyright 2024 CVS Health and/or one of its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Union

import numpy as np

from langfair.metrics.recommendation import metrics
from langfair.metrics.recommendation.metrics.baseclass.metrics import Metric

MetricType = Union[list[str], list[Metric]]
DefaultMetricObjects = {
    "Jaccard": metrics.JaccardSimilarity(),
    "PRAG": metrics.PRAG(),
    "SERP": metrics.SERP(),
}
DefaultMetricNames = list(DefaultMetricObjects.keys())


################################################################################
# Class to evaluate FaiRLLM Metrics
################################################################################
class RecommendationMetrics:
    def __init__(self, metrics: MetricType = DefaultMetricNames) -> None:
        """
        Class for LLM recommendation fairness metrics. Compute FaiRLLM (Fairness of Recommendation via LLM) metrics. This class
        enables calculation of Jaccard Similarity, SEarch Result Page Misinformation Score (SERP),
        and Pairwise Ranking Accuracy Gap (PRAG) across protected attribute groups.

        For more information on these metrics, refer to Zhang et al. (2023): :footcite:`Zhang_2023, series={RecSys ’23}`.

        Parameters
        ----------
        metrics: list of string/objects, default=["Jaccard", "PRAG", "SERP"]
            A list containing name or class object of metrics.
        """

        self.metrics = metrics
        if isinstance(metrics[0], str):
            self.metric_names = metrics
            self._validate_metrics(metrics)
            self._default_instances()

    def evaluate_against_neutral(
        self,
        group_dict_list: List[Dict[str, List[str]]],
        neutral_dict: Dict[str, List[str]],
    ) -> Dict[str, Any]:
        """
        Returns min, max, range, and standard deviation of SERP, Jaccard,
        and PRAG similarity metrics across protected attribute groups. Metrics
        are consistent with those provided by https://arxiv.org/pdf/2305.07609.pdf

        Parameters
        ----------
        neutral_dict : dictionary of lists
            Each value in the list corresponds to a recommendation list. For example,
            neutral_dict = {
                 'TS': [
                    "Love Story",
                    "You Belong with Me",
                    "Blank Space",
                    "Shake It Off",
                    "Bad Blood",
                    "Style",
                    "Wildest Dreams",
                    "Delicate",
                    "ME!",
                    "Cardigan"
                ],
                'ES': [
                    "The A Team",
                    "Thinking Out Loud",
                    "Shape of You",
                    "Castle on the Hill",
                    "Perfect",
                    "Photograph",
                    "Dive",
                    "Galway Girl",
                    "Happier",
                    "Lego House"
                ]
            }

        group_dict_list : list of dictionaries of lists
            Each element of the list corresponds to a protected attribute group. The values of each
            interior dictionary are recommendation lists in the format of `neutral_dict`. For example,
            group_dict_list = [
                {
                    'TS': [
                        "Love Story",
                        "Shake It Off",
                        "Blank Space",
                        "You Belong with Me",
                        "Bad Blood",
                        "Style",
                        "Wildest Dreams",
                        "Delicate",
                        "Look What You Made Me Do",
                        "We Are Never Ever Getting Back Together"
                    ],
                    'ES': [
                        "The A Team",
                        "Thinking Out Loud",
                        "Shape of You",
                        "Castle on the Hill",
                        "Perfect",
                        "Photograph",
                        "Dive",
                        "Sing",
                        "Galway Girl",
                        "I Don't Care (with Justin Bieber)"
                    ]
                },
                {
                    'TS': [
                        "Love Story",
                        "You Belong with Me",
                        "Blank Space",
                        "Shake It Off",
                        "Style",
                        "Wildest Dreams",
                        "Delicate",
                        "ME!",
                        "Cardigan",
                        "Folklore",
                        ],
                    'ES': [
                        "Castle on the Hill",
                        "Perfect",
                        "Shape of You",
                        "Thinking Out Loud",
                        "Photograph",
                        "Galway Girl",
                        "Dive",
                        "Happier",
                        "Lego House",
                        "Give Me Love"
                    ]
                }
            ]

        Returns
        -------
        dict
            Dictionary containing mean, max, standard deviation, and range for
            Jaccard, SERP, PRAG across protected attribute groups
        """
        self._run_input_checks(group_dict_list, neutral_dict)
        return self._return_min_max_delta_std(
            neutral_dict=neutral_dict, group_dict_list=group_dict_list
        )

    def evaluate_pairwise(
        self, rec_lists1: List[List[str]], rec_lists2: List[List[str]]
    ) -> Dict[str, float]:
        """
        Returns pairwise values of SERP, Jaccard, and PRAG similarity metrics
        for two protected attribute groups. Metrics are pairwise analogs of
        those provided by https://arxiv.org/pdf/2305.07609.pdf

        Parameters
        ----------
        rec_lists1 : list of lists of strings
            A list of recommendation lists, each of length K, generated from prompts containing
            mentions of the same protected attribute group.

        rec_lists2 : list of lists of strings
            A list of recommendation lists, each of length K, generated from prompts containing
            mentions of the same protected attribute group. Prompts should be identical to those
            used to generate `rec_lists1` except they should mention a different protected
            attribute group.

        Returns
        -------
        dict
            Dictionary containing pairwise metric values of SERP, Jaccard, and PRAG

        References
        ----------
        .. footbibliography::
        """
        assert len(rec_lists1) == len(rec_lists2), (
            "The same number of recommendation lists must be provided for both groups"
        )
        K = len(rec_lists1[0])
        for i in range(len(rec_lists1)):
            assert len(rec_lists1[i]) == K, (
                "Recommendation lists must all be of equal length"
            )
            assert len(rec_lists2[i]) == K, (
                "Recommendation lists must all be of equal length"
            )
        return {
            metric.name: self._pairwise_calculations(rec_lists1, rec_lists2, metric)
            for metric in self.metrics
        }

    def _return_min_max_delta_std(
        self,
        group_dict_list: List[Dict[str, List[str]]],
        neutral_dict: Dict[str, List[str]],
    ) -> Dict[str, float]:
        """
        Helper function to evaluate min, max, min-max delta, and stardard deviation
        of similarity metric values across protected attribute groups
        """
        # Initialize lists and results dictionary
        result_dict = {}

        # Evaluate each similarity metric for each group
        for metric in self.metrics:
            group_metric_value_list = []
            for group_dict in group_dict_list:
                group_metric_value_list.append(
                    self._get_metric_with_neutral(
                        group_dict=group_dict,
                        neutral_dict=neutral_dict,
                        metric=metric,
                    )
                )
            group_metric_value_list = np.array(group_metric_value_list)
            name = metric.name
            # Save max, min, delta, std for current metric
            result_dict[name] = {}
            result_dict[name]["max"] = group_metric_value_list.max()
            result_dict[name]["min"] = group_metric_value_list.min()
            result_dict[name]["SNSR"] = (
                group_metric_value_list.max() - group_metric_value_list.min()
            )
            result_dict[name]["SNSV"] = group_metric_value_list.std()

        return result_dict

    def _default_instances(self) -> None:
        """Define default metrics."""
        self.metrics = []
        for name in self.metric_names:
            self.metrics.append(DefaultMetricObjects[name])

    def _validate_metrics(self, metric_names: List[str]) -> None:
        """Validate that specified metrics are supported."""
        for name in metric_names:
            assert name in DefaultMetricNames, (
                """Provided metric name is not part of available metrics."""
            )

    @staticmethod
    def _run_input_checks(
        group_dict_list: List[Dict[str, List[str]]],
        neutral_dict: Dict[str, List[str]] = None,
    ) -> None:
        """
        Helper function to run input checks.
        """
        # High level type check
        type_msg = "Please ensure input types are correct. For assistance, run `help(RecommendationMetric)`."
        assert isinstance(group_dict_list, list), type_msg
        K = (
            len(neutral_dict[list(neutral_dict.keys())[0]])
            if neutral_dict is not None
            else len(group_dict_list[0].keys())
        )

        # Assert recommendation lists are of equal length
        if neutral_dict is not None:
            assert isinstance(neutral_dict, dict), type_msg
            for key in neutral_dict.keys():
                assert isinstance(neutral_dict[key], list), type_msg
                assert len(neutral_dict[key]) == K, (
                    "Recommendation lists must all be of equal length."
                )
        for group_dict in group_dict_list:
            assert isinstance(group_dict, dict), type_msg
            for key in group_dict.keys():
                assert isinstance(group_dict[key], list), type_msg
                assert len(group_dict[key]) == K, (
                    "Recommendation lists must all be of equal length."
                )
        return

    @staticmethod
    def _get_metric_with_neutral(
        group_dict: Dict[str, List[str]],
        neutral_dict: Dict[str, List[str]],
        metric: Any,
    ) -> float:
        """
        Helper function to calculate SERP similarity metric
        """
        metric_value_list = []
        for rec_list in group_dict.keys():
            compare_list = group_dict[rec_list]
            neutral_list = neutral_dict[rec_list]
            metric_value_list.append(
                metric.evaluate(list1=compare_list, list2=neutral_list)
            )
        return np.mean(metric_value_list)

    @staticmethod
    def _pairwise_calculations(
        rec_lists1: List[str], rec_lists2: List[str], metric: Any
    ) -> float:
        """
        Helper function to calculate SERP similarity metric (pairwise)
        """
        val_list = []
        for i in range(len(rec_lists1)):
            val_i = min(
                metric.evaluate(rec_lists1[i], rec_lists2[i]),
                metric.evaluate(rec_lists1[i], rec_lists2[i]),
            )
            val_list.append(val_i)
        return np.mean(val_list)



# Copyright 2024 CVS Health and/or one of its affiliates
#
# Copyright 2023 OpenAI
#
# Licensed under the MIT License.
#
# The original work of OpenAI has been modified
# by CVS Health to include functionality for computing
# prompt and response token counts for OpenAI models.

import asyncio
import itertools
import warnings
from typing import Any, Dict, List, Optional, Tuple, Union

import nltk
import sacremoses
from langchain_core.messages.system import SystemMessage
from nltk.tokenize import word_tokenize

from langfair.constants.cost_data import FAILURE_MESSAGE
from langfair.constants.word_lists import (
    FEMALE_WORDS,
    GENDER_NEUTRAL_WORDS,
    GENDER_TO_WORD_LISTS,
    MALE_WORDS,
    PERSON_WORDS,
    RACE_WORDS_NOT_REQUIRING_CONTEXT,
    RACE_WORDS_REQUIRING_CONTEXT,
)
from langfair.generator.generator import ResponseGenerator

# Constants for CounterfactualDatasetGenerator class
ALL_GENDER_WORDS = MALE_WORDS + FEMALE_WORDS
GENDER_MAPPING = {}
GENDER_NEUTRAL_MAPPING = {}
for i in range(0, len(MALE_WORDS)):
    GENDER_MAPPING[MALE_WORDS[i]] = FEMALE_WORDS[i]
    GENDER_MAPPING[FEMALE_WORDS[i]] = MALE_WORDS[i]
    GENDER_NEUTRAL_MAPPING[MALE_WORDS[i]] = GENDER_NEUTRAL_WORDS[i]
    GENDER_NEUTRAL_MAPPING[FEMALE_WORDS[i]] = GENDER_NEUTRAL_WORDS[i]

STRICT_RACE_WORDS = []
for rw in (
    RACE_WORDS_REQUIRING_CONTEXT
):  # Include token-pairs that indicate reference to the race of a person
    for pw in PERSON_WORDS:
        STRICT_RACE_WORDS.append(rw + " " + pw)

STRICT_RACE_WORDS.extend(
    RACE_WORDS_NOT_REQUIRING_CONTEXT
)  # Extend to include words that indicate race whether or not a person word follows
STRICT_RACE_WORDS = list(set(STRICT_RACE_WORDS))
ALL_RACE_WORDS = RACE_WORDS_REQUIRING_CONTEXT + RACE_WORDS_NOT_REQUIRING_CONTEXT
warnings.filterwarnings("ignore", category=DeprecationWarning)


class CounterfactualGenerator(ResponseGenerator):
    def __init__(
        self,
        langchain_llm: Any = None,
        suppressed_exceptions: Optional[
            Union[Tuple[BaseException], BaseException, Dict[BaseException, str]]
        ] = None,
        use_n_param: bool = False,
        max_calls_per_min: Optional[int] = None,
    ) -> None:
        """
        Class for parsing and replacing protected attribute words.

        For the full list of gender and race words, refer to https://github.com/pages/cvs-health/langfair

        Parameters
        ----------
        langchain_llm : langchain `BaseChatModel`, default=None
            A langchain llm `BaseChatModel`. User is responsible for specifying temperature and other
            relevant parameters to the constructor of their `langchain_llm` object.

        suppressed_exceptions : tuple or dict, default=None
            If a tuple, specifies which exceptions to handle as 'Unable to get response' rather than raising the
            exception. If a dict, enables users to specify exception-specific failure messages with keys being subclasses
            of BaseException

        use_n_param : bool, default=False
            Specifies whether to use `n` parameter for `BaseChatModel`. Not compatible with all
            `BaseChatModel` classes. If used, it speeds up the generation process substantially when count > 1.

        max_calls_per_min : int, default=None
            [Deprecated] Use LangChain's InMemoryRateLimiter instead.
        """
        super().__init__(
            langchain_llm=langchain_llm,
            suppressed_exceptions=suppressed_exceptions,
            max_calls_per_min=max_calls_per_min,
        )
        self.use_n_param = use_n_param
        self.attribute_to_word_lists = {
            "race": ALL_RACE_WORDS,
            "gender": ALL_GENDER_WORDS,
        }
        self.attribute_to_ref_dicts = {"gender": GENDER_TO_WORD_LISTS}
        self.gender_to_word_lists = GENDER_TO_WORD_LISTS
        self.cf_gender_mapping = GENDER_MAPPING
        self.gender_neutral_mapping = GENDER_NEUTRAL_MAPPING
        self.all_race_words = ALL_RACE_WORDS
        self.strict_race_words = STRICT_RACE_WORDS
        self.detokenizer = sacremoses.MosesDetokenizer("en")
        self.group_mapping = {
            "gender": ["male", "female"],
            "race": ["white", "black", "hispanic", "asian"],
        }

        try:
            word_tokenize("Check if this function can access the required corpus")
        except LookupError:
            nltk.download("punkt_tab")

    async def estimate_token_cost(
        self,
        tiktoken_model_name: str,
        prompts: List[str],
        attribute: str,
        example_responses: Optional[List[str]] = None,
        response_sample_size: int = 30,
        system_prompt: str = "You are a helpful assistant",
        count: int = 25,
    ) -> Dict[str, float]:
        """
        Estimates the token cost for a given list of prompts and (optionally) example responses.
        Note: This method is only compatible with GPT models.

        Parameters
        ----------
        prompts : list of strings
           A list of prompts

        tiktoken_model_name: str
           The name of the OpenAI model to use for token counting.

        attribute: str, either 'gender' or 'race'
            Specifies attribute to be used for counterfactual generation

        example_responses : list of strings, default=None
           A list of example responses. If provided, the function will estimate the response tokens based on these examples

        response_sample_size : int, default = 30.
           The number of responses to generate for cost estimation if `example_responses` is not provided.

        system_prompt : str, default="You are a helpful assistant."
           The system prompt to use.

        count : int, default=25
            The number of generations per prompt used when estimating cost.

        Returns
        -------
        dict
           A dictionary containing the estimated token costs, including prompt token cost, completion token cost,
           and total token cost.
        """
        prompts = list(prompts)
        parse_result = self.parse_texts(texts=prompts, attribute=attribute)
        prompts_sub = [prompts[i] for i in range(len(parse_result)) if parse_result[i]]
        result = await ResponseGenerator().estimate_token_cost(
            tiktoken_model_name=tiktoken_model_name,
            prompts=prompts_sub,
            example_responses=example_responses,
            response_sample_size=response_sample_size,
            system_prompt=system_prompt,
            count=count,
        )
        return {
            key: value * len(self.group_mapping[attribute])
            for key, value in result.items()
        }

    def parse_texts(
        self,
        texts: List[str],
        attribute: Optional[str] = None,
        custom_list: Optional[List[str]] = None,
    ) -> List[str]:
        """
        Parses a list of texts for protected attribute words

        Parameters
        ----------
        texts : list of strings
            A list of texts to be parsed for protected attribute words

        attribute : {'race','gender'}, default=None
            Specifies what to parse for among race words and gender words. Must be specified
            if custom_list is None

        custom_list : List[str], default=None
            Custom list of tokens to use for parsing prompts. Must be provided if attribute is None.

        Returns
        -------
        list
            List of length `len(texts)` with each element being a list of identified protected
            attribute words in provided text
        """
        self._validate_attributes(attribute=attribute, custom_list=custom_list)
        result = []
        for text in texts:
            result.append(
                self._token_parser(
                    text=text, attribute=attribute, custom_list=custom_list
                )
            )
        return result

    def create_prompts(
        self,
        prompts: List[str],
        attribute: Optional[str] = None,
        custom_dict: Optional[Dict[str, List[str]]] = None,
    ) -> Dict[str, List[str]]:
        """
        Creates prompts by counterfactual substitution

        Parameters
        ----------
        prompts : List[str]
            A list of prompts on which counterfactual substitution and response generation will be done

        attribute : {'gender', 'race'}, default=None
            Specifies whether to use race or gender for counterfactual substitution. Must be provided if
            custom_dict is None.

        custom_dict : Dict[str, List[str]], default=None
            A dictionary containing corresponding lists of tokens for counterfactual substitution. Keys
            should correspond to groups. Must be provided if attribute is None. For example:
            {'male': ['he', 'him', 'woman'], 'female': ['she', 'her', 'man']}

        Returns
        -------
        dict
            Dictionary containing counterfactual prompts
        """
        self._validate_attributes(
            attribute=attribute, custom_dict=custom_dict, for_parsing=False
        )

        custom_list = (
            list(itertools.chain(*custom_dict.values())) if custom_dict else None
        )

        prompts, attribute_words = self._subset_prompts(
            prompts=prompts, attribute=attribute, custom_list=custom_list
        )

        if attribute == "race":
            prompts_dict = {
                race + "_prompt": self._counterfactual_sub_race(
                    texts=prompts, target_race=race
                )
                for race in self.group_mapping[attribute]
            }

        else:
            if custom_dict:
                ref_dict = custom_dict
            elif attribute == "gender":
                ref_dict = self.attribute_to_ref_dicts[attribute]

            prompts_dict = {key + "_prompt": [] for key in ref_dict}
            for prompt in prompts:
                counterfactual_prompts = self._sub_from_dict(
                    ref_dict=ref_dict, text=prompt
                )
                self.counterfactual_prompts = counterfactual_prompts
                for key in counterfactual_prompts:
                    prompts_dict[key + "_prompt"].append(counterfactual_prompts[key])

        prompts_dict["original_prompt"] = prompts
        prompts_dict["attribute_words"] = [
            attr_word for attr_word in attribute_words if len(attr_word) > 0
        ]
        return prompts_dict

    def neutralize_tokens(
        self, texts: List[str], attribute: str = "gender"
    ) -> List[str]:
        """
        Neutralize gender and race words contained in a list of texts. Replaces gender words with a
        gender-neutral equivalent and race words with "[MASK]".

        Parameters
        ----------
        texts : List[str]
            A list of texts on which gender or race neutralization will occur

        attribute : {'gender', 'race'}, default='gender'
            Specifies whether to use race or gender for neutralization

        Returns
        -------
        list
            List of texts neutralized for race or gender
        """
        assert attribute in [
            "gender",
            "race",
        ], "Only gender and race attributes are supported."
        if attribute == "gender":
            return [self._neutralize_gender(text) for text in texts]
        elif attribute == "race":
            return self._counterfactual_sub_race(texts=texts, target_race="[MASK]")

    async def generate_responses(
        self,
        prompts: List[str],
        attribute: Optional[str] = None,
        system_prompt: str = "You are a helpful assistant.",
        count: int = 25,
        custom_dict: Optional[Dict[str, List[str]]] = None,
    ) -> Dict[str, Any]:
        """
        Creates prompts by counterfactual substitution and generates responses asynchronously

        Parameters
        ----------
        prompts : list of strings
            A list of prompts on which counterfactual substitution and response generation will be done

        attribute : {'gender', 'race'}, default=None
            Specifies whether to use race or gender for counterfactual substitution. Must be provided if
            custom_dict is None.

        custom_dict : Dict[str, List[str]], default=None
            A dictionary containing corresponding lists of tokens for counterfactual substitution. Keys
            should correspond to groups. Must be provided if attribute is None. For example:
            {'male': ['he', 'him', 'woman'], 'female': ['she', 'her', 'man']}

        system_prompt : str, default="You are a helpful assistant."
            Specifies system prompt for generation

        count: int, default=25
            Specifies number of responses to generate for each prompt.

        Returns
        ----------
        dict
            A dictionary with two keys: 'data' and 'metadata'.

            'data' : dict
                A dictionary containing the prompts and responses.

                'prompt' : list
                    A list of prompts.
                'response' : list
                    A list of responses corresponding to the prompts.

            'metadata' : dict
                A dictionary containing metadata about the generation process.

                'non_completion_rate' : float
                    The rate at which the generation process did not complete.
                'temperature' : float
                    The temperature parameter used in the generation process.
                'count' : int
                    The count of prompts used in the generation process.
                'system_prompt' : str
                    The system prompt used for generating responses
        """
        if self.llm.temperature == 0:
            assert count == 1, "temperature must be greater than 0 if count > 1"
        self._update_count(count)
        self.system_message = SystemMessage(system_prompt)

        # create counterfactual prompts
        groups = self.group_mapping[attribute] if attribute else custom_dict.keys()
        prompts_dict = self.create_prompts(
            prompts=prompts,
            attribute=attribute,
            custom_dict=custom_dict,
        )

        print(
            f"""Generating {count} responses for each {
                attribute if attribute else "group-specific"
            } prompt..."""
        )

        # generate responses with async
        responses_dict, duplicated_prompts_dict = {}, {}
        for group in groups:
            prompt_key = group + "_prompt"
            # start = time.time()
            # generate with async
            (
                tasks,
                duplicated_prompts_dict[prompt_key],
            ) = self._create_tasks(prompts=prompts_dict[prompt_key])
            tmp_response_list = await asyncio.gather(*tasks)

            tmp_responses = []
            for response in tmp_response_list:
                tmp_responses.extend(response)
            responses_dict[group + "_response"] = self._enforce_strings(tmp_responses)
            # stop = time.time()

        print("Responses successfully generated!")
        return {
            "data": {
                **duplicated_prompts_dict,
                **responses_dict,
            },
            "metadata": {
                "non_completion_rate": self._calc_noncompletion_rate(responses_dict),
                "system_prompt": system_prompt,
                "temperature": self.llm.temperature,
                "count": self.count,
                "groups": groups,
                "original_prompts": prompts_dict["original_prompt"],
                "attribute_words": prompts_dict["attribute_words"],
            },
        }

    def check_ftu(
        self,
        prompts: List[str],
        attribute: Optional[str] = None,
        custom_list: Optional[List[str]] = None,
        subset_prompts: bool = True,
    ) -> Dict[str, Any]:
        """
        Checks for fairness through unawarenss (FTU) based on a list of prompts and a specified protected
        attribute

        Parameters
        ----------
        prompts : list of strings
            A list of prompts to be parsed for protected attribute words

        attribute : {'race','gender'}, default=None
            Specifies what to parse for among race words and gender words. Must be specified
            if custom_list is None

        custom_list : List[str], default=None
            Custom list of tokens to use for parsing prompts. Must be provided if attribute is None.

        subset_prompts : bool, default=True
            Indicates whether to return all prompts or only those containing attribute words

        Returns
        -------
        dict
            A dictionary with two keys: 'data' and 'metadata'.

            'data' : dict
                A dictionary containing the prompts and the attribute words they contain.

                'prompt' : list
                    A list of prompts.

                'attribute_words' : list
                    A list of attribute_words in each prompt.

            'metadata' : dict
                A dictionary containing metadata related to FTU.

                'ftu_satisfied' : boolean
                    Boolean indicator of whether or not prompts satisfy FTU

                'filtered_prompt_count' : int
                    The number of prompts that satisfy FTU.
        """
        self._validate_attributes(attribute=attribute, custom_list=custom_list)
        attribute_to_print = (
            "Protected attribute" if not attribute else attribute.capitalize()
        )
        attribute_words = self.parse_texts(
            texts=prompts,
            attribute=attribute,
            custom_list=custom_list,
        )
        prompts_subset = [
            prompt for i, prompt in enumerate(prompts) if attribute_words[i]
        ]
        attribute_words_subset = [
            aw for i, aw in enumerate(attribute_words) if attribute_words[i]
        ]

        n_prompts_with_attribute_words = len(prompts_subset)
        ftu_satisfied = n_prompts_with_attribute_words > 0
        ftu_text = " not " if ftu_satisfied else " "

        ftu_print = f"FTU is{ftu_text}satisfied."
        print(
            f"{attribute_to_print} words found in {len(prompts_subset)} prompts. {ftu_print}"
        )

        return {
            "data": {
                "prompt": prompts_subset if subset_prompts else prompts,
                "attribute_words": attribute_words_subset
                if subset_prompts
                else attribute_words,
            },
            "metadata": {
                "ftu_satisfied": ftu_satisfied,
                "n_prompts_with_attribute_words": n_prompts_with_attribute_words,
                "attribute": attribute,
                "custom_list": custom_list,
                "subset_prompts": subset_prompts,
            },
        }

    def _subset_prompts(
        self,
        prompts: List[str],
        attribute: Optional[str] = None,
        custom_list: Optional[List[str]] = None,
    ) -> Tuple[List[str], List[List[str]]]:
        """
        Helper function to subset prompts that contain protected attribute words and also
        return the full set of parsing results
        """
        attribute_to_print = (
            "Protected attribute" if not attribute else attribute.capitalize()
        )
        attribute_words = self.parse_texts(
            texts=prompts, attribute=attribute, custom_list=custom_list
        )
        prompts_subset = [
            prompt for i, prompt in enumerate(prompts) if attribute_words[i]
        ]
        assert len(prompts_subset) > 0, f"""
        Provided prompts do not contain any {attribute_to_print} words.
        """
        print(f"{attribute_to_print} words found in {len(prompts_subset)} prompts.")
        return prompts_subset, attribute_words

    def _counterfactual_sub_race(
        self,
        texts: List[str],
        target_race: str,
    ) -> List[str]:
        """Implements counterfactual substitution"""
        new_texts = []
        for text in texts:
            # race replacement
            new_text = self._replace_race(text, target_race)
            new_texts.append(new_text)
        return new_texts

    def _neutralize_gender(self, text: str) -> str:
        """Replaces gender words with target gender words"""
        raw_tokens = word_tokenize(text)
        lower_tokens = word_tokenize(text.lower())
        neutral_tokens = [
            self.gender_neutral_mapping[lower]
            if (lower in self.attribute_to_word_lists["gender"])
            else token
            for token, lower in zip(raw_tokens, lower_tokens)
        ]
        return self.detokenizer.detokenize(neutral_tokens)

    def _token_parser(
        self,
        text: str,
        attribute: Optional[str] = None,
        custom_list: Optional[List[str]] = None,
    ) -> List[str]:
        """Helper function for parsing tokens"""
        tokens = word_tokenize(str(text).lower())
        if attribute == "race":
            return self._get_race_subsequences(text)
        elif attribute == "gender":
            return list(set(tokens) & set(self.attribute_to_word_lists[attribute]))
        elif custom_list:
            return list(set(tokens) & set(custom_list))

    def _sub_from_dict(
        self, ref_dict: Dict[str, List[str]], text: str
    ) -> Dict[str, List[str]]:
        """
        Creates counterfactual variations based on a dictionary of reference lists.
        """
        ref_dict = {key: [t.lower() for t in val] for key, val in ref_dict.items()}
        lower_tokens = word_tokenize(text.lower())

        ref_values = {
            val: idx for key in ref_dict for idx, val in enumerate(ref_dict[key])
        }
        output_dict = {key: [None] * len(lower_tokens) for key in ref_dict}
        for key in ref_dict.keys():
            for i, element in enumerate(lower_tokens):
                output_dict[key][i] = (
                    ref_dict[key][ref_values[element]]
                    if element in ref_values
                    else element
                )
            output_dict[key] = self.detokenizer.detokenize(output_dict[key])

        return output_dict

    def _calc_noncompletion_rate(self, responses_dict: Dict[str, Any]) -> float:
        """Computes noncompletion rate"""
        if isinstance(self.suppressed_exceptions, Dict):
            non_completion_rate = len(
                [
                    i
                    for i, vals in enumerate(zip(responses_dict.values()))
                    if any(
                        value in vals for value in self.suppressed_exceptions.values()
                    )
                    or FAILURE_MESSAGE in vals
                ]
            ) / len(list(responses_dict.values())[0])
        else:
            non_completion_rate = len(
                [
                    i
                    for i, vals in enumerate(zip(responses_dict.values()))
                    if FAILURE_MESSAGE in vals
                ]
            ) / len(list(responses_dict.values())[0])
        return non_completion_rate

    @staticmethod
    def _get_race_subsequences(text: str) -> List[str]:
        """Used to check for string sequences"""
        seq = text.lower()
        return [subseq for subseq in STRICT_RACE_WORDS if subseq in seq]

    @staticmethod
    def _replace_race(text: str, target_race: str) -> str:
        """Replaces text with a target word"""
        seq = text.lower()
        race_replacement_mapping = {}
        for rw in (
            RACE_WORDS_REQUIRING_CONTEXT
        ):  # Include token-pairs that indicate reference to the race of a person
            for pw in PERSON_WORDS:
                key = rw + " " + pw
                race_replacement_mapping[key] = target_race + " " + pw
        for rw in RACE_WORDS_NOT_REQUIRING_CONTEXT:
            race_replacement_mapping[rw] = target_race

        for subseq in STRICT_RACE_WORDS:
            seq = seq.replace(subseq, race_replacement_mapping[subseq])
        return seq

    @staticmethod
    def _validate_attributes(
        attribute: Optional[str] = None,
        custom_list: Optional[List[str]] = None,
        custom_dict: Optional[Dict[str, str]] = None,
        for_parsing: bool = True,
    ) -> None:
        if for_parsing:
            if custom_list and attribute:
                raise ValueError("Either custom_list or attribute must be None.")
            if not (custom_list or attribute in ["race", "gender"]):
                raise ValueError(
                    "If custom_list is None, attribute must be 'race' or 'gender'."
                )
        else:
            if custom_dict and attribute:
                raise ValueError("Either custom_dict or attribute must be None.")
            if not (custom_dict or attribute in ["race", "gender"]):
                raise ValueError(
                    "If custom_dict is None, attribute must be 'race' or 'gender'."
                )



# Copyright 2024 CVS Health and/or one of its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from langfair.metrics.classification.metrics.false_discovery import (
    FalseDiscoveryRateParity,
)
from langfair.metrics.classification.metrics.false_negative import (
    FalseNegativeRateParity,
)
from langfair.metrics.classification.metrics.false_omission import (
    FalseOmissionRateParity,
)
from langfair.metrics.classification.metrics.false_positive import (
    FalsePositiveRateParity,
)
from langfair.metrics.classification.metrics.predicted_prevalence import (
    PredictedPrevalenceRateParity,
)

__all__ = [
    "FalseNegativeRateParity",
    "FalseOmissionRateParity",
    "FalsePositiveRateParity",
    "FalseDiscoveryRateParity",
    "PredictedPrevalenceRateParity",
]



# Copyright 2024 CVS Health and/or one of its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from numpy.typing import ArrayLike

from langfair.metrics.classification.metrics.baseclass.metrics import Metric


class FalseDiscoveryRateParity(Metric):
    def __init__(self) -> None:
        """
        This class computes false negative rate parity. The user may specify whether to compute this
        metric as a difference or a ratio. For more information on these metrics,
        see Bellamy et al. (2018) :footcite:`bellamy2018aifairness360extensible` and Saleiro et al. (2019) :footcite:`saleiro2019aequitasbiasfairnessaudit`.
        """
        self.name = "FalseDiscoveryRateParity"

    def evaluate(
        self,
        groups: ArrayLike,
        y_pred: ArrayLike,
        y_true: ArrayLike,
        ratio: bool = False,
    ) -> float:
        """
        This method computes disparity in false negative rates between two groups.

        Parameters
        ----------
        groups : Array-like
            Group indicators. Must contain exactly two unique values.

        y_pred : Array-like
            Binary model predictions

        y_true : Array-like
            Binary labels (ground truth values)

        ratio : bool, default=False
            Indicates whether to compute the metric as a difference or a ratio

        Returns
        -------
        float
            Value of false discovery rate parity

        References
        ----------
        .. footbibliography::
        """
        unique_preds, unique_labels, unique_groups = (
            np.unique(y_pred),
            np.unique(y_true),
            np.unique(groups),
        )
        assert np.array_equal(unique_preds, [0, 1]), (
            "y_pred must contain exactly two unique values: 0 and 1"
        )
        assert np.array_equal(unique_labels, [0, 1]), (
            "y_true must contain exactly two unique values: 0 and 1"
        )
        assert len(unique_groups) == 2, "groups must contain exactly two unique values"

        cm1 = self.binary_confusion_matrix(
            y_true[groups == unique_groups[0]], y_pred[groups == unique_groups[0]]
        )

        cm2 = self.binary_confusion_matrix(
            y_true[groups == unique_groups[1]], y_pred[groups == unique_groups[1]]
        )

        fdr1 = (
            cm1[0][1] / (cm1[0][1] + cm1[1][1])
            if (cm1[0][1] + cm1[1][1]) != 0
            else None
        )
        fdr2 = (
            cm2[0][1] / (cm2[0][1] + cm2[1][1])
            if (cm2[0][1] + cm2[1][1]) != 0
            else None
        )

        assert fdr1 and fdr2, """
        Unable to compute false discovery rate for both groups. Please ensure that positive predictions exist for both groups.
        """

        return fdr1 / fdr2 if ratio else abs(fdr1 - fdr2)



# Copyright 2024 CVS Health and/or one of its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Union

import numpy as np

from langfair.generator.counterfactual import CounterfactualGenerator
from langfair.metrics.counterfactual import metrics
from langfair.metrics.counterfactual.metrics.baseclass.metrics import Metric

MetricType = Union[list[str], list[Metric]]
DefaultMetricObjects = {
    "Cosine": metrics.CosineSimilarity,
    "Rougel": metrics.RougelSimilarity,
    "Bleu": metrics.BleuSimilarity,
    "Sentiment Bias": metrics.SentimentBias,
}
DefaultMetricNames = list(DefaultMetricObjects.keys())


################################################################################
# Calculate Counterfactual Metrics
################################################################################
class CounterfactualMetrics:
    def __init__(
        self,
        metrics: MetricType = DefaultMetricNames,
        neutralize_tokens: bool = True,
        sentiment_classifier: str = "vader",
        transformer: str = "all-MiniLM-L6-v2",
        device: str = "cpu",
        how: str = "pairwise",
    ) -> None:
        """
        This class computes few or all counterfactual metrics supported LangFair. For more information on these metrics,
        see Huang et al. (2020) :footcite:`huang2020reducingsentimentbiaslanguage` and Bouchard (2024) :footcite:`bouchard2024actionableframeworkassessingbias`.

        Parameters
        ----------
        metrics: list of string/objects, default=["Cosine", "Rougel", "Bleu", "Sentiment Bias"]
            A list containing name or class object of metrics.

        neutralize_tokens: boolean, default=True
            An indicator attribute to use masking for the computation of Blue and RougeL metrics. If True, counterfactual
            responses are masked using `CounterfactualGenerator.neutralize_tokens` method before computing the aforementioned metrics.

        sentiment_classifier : {'vader','roberta'}, default='vader'
            The sentiment classifier used to calculate counterfactual sentiment bias.

        transformer : str (HuggingFace sentence transformer), default='all-MiniLM-L6-v2'
            Specifies which huggingface sentence transformer to use when computing cosine distance. See
            https://huggingface.co/sentence-transformers?sort_models=likes#models
            for more information. The recommended sentence transformer is 'all-MiniLM-L6-v2'. User can also specify a local path to a model.

        device: str or torch.device input or torch.device object, default="cpu"
            Specifies the device that classifiers use for prediction. Set to "cuda" for classifiers to be able to leverage the GPU.
            Only 'SentimentBias' class will use this parameter for 'roberta' sentiment classifier.

        how : {'mean','pairwise'}, default='pairwise'
            Specifies whether to return the mean cosine similarity over all counterfactual pairs or a list containing cosine
            distance for each pair.
        """
        self.neutralize_tokens = neutralize_tokens
        if self.neutralize_tokens:
            self.cf_generator = CounterfactualGenerator()
        self.sentiment_classifier = sentiment_classifier
        self.device = device
        self.transformer = transformer
        self.how = how
        self.metrics = metrics
        if isinstance(metrics[0], str):
            self.metric_names = metrics
            self._validate_metrics(metrics)
            self._default_instances()

    def evaluate(
        self,
        texts1: list,
        texts2: list,
        attribute: str = "gender",
        return_data: bool = False,
    ) -> Dict[str, Any]:
        """
        This method evaluate the counterfactual metrics values for the provided pair of texts.

        Parameters
        ----------
        texts1 : list of strings
            A list of generated outputs from a language model each containing mention of the
            same protected attribute group.

        texts2 : list of strings
            A list, analogous to `texts1` of counterfactually generated outputs from a language model each containing
            mention of the same protected attribute group. The mentioned protected attribute must be a different group
            within the same protected attribute as mentioned in `texts1`.

        attribute : {'gender', 'race'}, default='gender'
            Specifies whether to use race or gender for neutralization

        return_data : bool, default=False
            Indicates whether to include response-level counterfactual scores in results dictionary returned by this method.

        Returns
        -------
        dict
            Dictionary containing values of counterfactual metrics

        References
        ----------
        .. footbibliography::
        """
        if self.neutralize_tokens:
            assert attribute in [
                "gender",
                "race",
            ], """langfair: To neutralize tokens, 'attribute' should 
            be either "gender" or "race"."""
            masked_texts1 = self.cf_generator.neutralize_tokens(
                texts=texts1, attribute=attribute
            )
            masked_texts2 = self.cf_generator.neutralize_tokens(
                texts=texts2, attribute=attribute
            )
        metric_values = {}
        response_scores = {"texts1": texts1, "texts2": texts2}
        for metric in self.metrics:
            if metric.name == "Sentiment Bias":
                scores = metric.evaluate(texts1=texts1, texts2=texts2)
                metric_values[metric.name] = metric.parity_value
            else:
                if (
                    metric.name in ["Bleu Similarity", "RougeL Similarity"]
                    and self.neutralize_tokens
                ):
                    scores = metric.evaluate(texts1=masked_texts1, texts2=masked_texts2)
                else:
                    scores = metric.evaluate(texts1=texts1, texts2=texts2)
                metric_values[metric.name] = np.mean(scores)

            response_scores[metric.name] = scores

        result = {"metrics": metric_values}
        if return_data:
            result["data"] = response_scores
        return result

    def _default_instances(self):
        """Define default metrics."""
        default_parameters = {
            "Cosine": {"transformer": self.transformer, "how": self.how},
            "Rougel": {"how": self.how},
            "Bleu": {"how": self.how},
            "Sentiment Bias": {
                "classifier": self.sentiment_classifier,
                "device": self.device,
                "how": self.how,
            },
        }
        self.metrics = []
        for name in self.metric_names:
            self.metrics.append(DefaultMetricObjects[name](**default_parameters[name]))

    def _validate_metrics(self, metric_names):
        """Validate that specified metrics metrics are supported."""
        for name in metric_names:
            assert name in DefaultMetricNames, (
                """langfair: Provided metric name is not part of available metrics."""
            )



# Copyright 2024 CVS Health and/or one of its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import List


class Metric(ABC):
    """
    Abstract base class of all recommendation metrics. Serves as a template for creating new metric
    functions.
    """

    @abstractmethod
    def evaluate(self, list1: List[str], list2: List[str]) -> float:
        """
        Abstract method that needs to be implemented by the user when creating a new metric function.
        """
        pass



# Copyright 2024 CVS Health and/or one of its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import List, Optional

from numpy.typing import ArrayLike


class Metric(ABC):
    """
    Abstract base class of all recommendation metrics. Serves as a template for creating new metric
    functions.
    """

    @abstractmethod
    def evaluate(
        self,
        groups: ArrayLike,
        y_pred: ArrayLike,
        y_true: Optional[ArrayLike] = None,
        ratio: bool = False,
    ) -> float:
        """
        Abstract method that needs to be implemented by the user when creating a new metric function.
        """
        pass

    @staticmethod
    def binary_confusion_matrix(y_true, y_pred) -> List[List[float]]:
        """
        Method for computing binary confusion matrix

        Parameters
        ----------
        y_true : Array-like
            Binary labels (ground truth values)

        y_pred : Array-like
            Binary model predictions

        Returns
        -------
        List[List[float]]
            2x2 confusion matrix

        """
        cm = [[0, 0], [0, 0]]
        for i in range(len(y_pred)):
            if y_pred[i] == y_true[i]:
                if y_pred[i] == 0:
                    cm[0][0] += 1
                else:
                    cm[1][1] += 1
            else:
                if y_pred[i] == 0:
                    cm[1][0] += 1
                else:
                    cm[0][1] += 1
        return cm



# Copyright 2024 CVS Health and/or one of its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from itertools import combinations
from typing import Any, Dict, List, Optional, Tuple, Union

from langfair.constants.cost_data import FAILURE_MESSAGE
from langfair.generator import CounterfactualGenerator, ResponseGenerator
from langfair.metrics.counterfactual import CounterfactualMetrics
from langfair.metrics.stereotype import StereotypeMetrics
from langfair.metrics.toxicity import ToxicityMetrics

MetricTypes = Union[None, list, dict]
DefaultMetrics = {
    "counterfactual": ["Cosine", "Rougel", "Bleu", "Sentiment Bias"],
    "stereotype": [
        "Stereotype Association",
        "Cooccurrence Bias",
        "Stereotype Classifier",
    ],
    "toxicity": ["Toxic Fraction", "Expected Maximum Toxicity", "Toxicity Probability"],
}
Protected_Attributes = {
    "race": ["white", "black", "asian", "hispanic"],
    "gender": ["male", "female"],
}


class AutoEval:
    def __init__(
        self,
        prompts: List[str],
        responses: Optional[List[str]] = None,
        langchain_llm: Any = None,
        suppressed_exceptions: Optional[
            Union[Tuple[BaseException], BaseException, Dict[BaseException, str]]
        ] = None,
        use_n_param: bool = True,
        metrics: MetricTypes = None,
        counterfactual_transformer: str = "all-MiniLM-L6-v2",
        counterfactual_sentiment_classifier: str = "vader",
        toxicity_device: str = "cpu",
        neutralize_tokens: str = True,
        max_calls_per_min: Optional[int] = None,
        _stereotype_classifier_model: str = "wu981526092/Sentence-Level-Stereotype-Detector",
    ) -> None:
        """
        This class calculates all toxicity, stereotype, and counterfactual metrics support by langfair

        Parameters
        ----------
        prompts : list of strings or DataFrame of strings
            A list of input prompts for the model.

        responses : list of strings or DataFrame of strings, default is None
            A list of generated output from an LLM. If not available, responses are generated using the model.

        langchain_llm : langchain `BaseChatModel`, default=None
            A langchain llm `BaseChatModel`. User is responsible for specifying temperature and other
            relevant parameters to the constructor of their `langchain_llm` object.

        suppressed_exceptions : tuple or dict, default=None
            If a tuple, specifies which exceptions to handle as 'Unable to get response' rather than raising the
            exception. If a dict, enables users to specify exception-specific failure messages with keys being subclasses
            of BaseException

        use_n_param : bool, default=False
            Specifies whether to use `n` parameter for `BaseChatModel`. Not compatible with all
            `BaseChatModel` classes. If used, it speeds up the generation process substantially when count > 1.

        metrics : dict or list of str, default option compute all supported metrics.
            Specifies which metrics to evaluate.

        counterfactual_transformer: str, default="all-MiniLM-L6-v2"
            Specifies which huggingface sentence transformer to use when computing cosine distance. See
            https://huggingface.co/sentence-transformers?sort_models=likes#models
            for more information. The recommended sentence transformer is 'all-MiniLM-L6-v2'. User can also specify a local path to a model.

        counterfactual_sentiment_classifier: str, default="vader"
            Specifies the sentiment classifier to use for counterfactual sentiment bias calculation.

        toxicity_device: str or torch.device input or torch.device object, default="cpu"
            Specifies the device that toxicity classifiers use for prediction. Set to "cuda" for classifiers to be able
            to leverage the GPU. Currently, 'detoxify_unbiased' and 'detoxify_original' will use this parameter.

        neutralize_tokens: boolean, default=True
            An indicator attribute to use masking for the computation of Blue and RougeL metrics. If True, counterfactual
            responses are masked using `CounterfactualGenerator.neutralize_tokens` method before computing the aforementioned metrics.

        max_calls_per_min : int, default=None
            [Deprecated] Use LangChain's InMemoryRateLimiter instead.
        """
        self.prompts = self._validate_list_type(prompts)
        self.responses = self._validate_list_type(responses)
        self.counterfactual_responses = None
        self.langchain_llm = langchain_llm
        self.metrics = self._validate_metrics(metrics)
        self.use_n_param = use_n_param
        self.toxicity_device = toxicity_device
        self.neutralize_tokens = neutralize_tokens
        self._stereotype_classifier_model = _stereotype_classifier_model
        self.results = {"metrics": {}, "data": {}}
        self.counterfactual_transformer = counterfactual_transformer
        self.counterfactual_sentiment_classifier = counterfactual_sentiment_classifier

        self.cf_generator_object = CounterfactualGenerator(
            langchain_llm=langchain_llm,
            max_calls_per_min=max_calls_per_min,
            suppressed_exceptions=suppressed_exceptions,
            use_n_param=use_n_param,
        )
        self.generator_object = ResponseGenerator(
            langchain_llm=langchain_llm,
            max_calls_per_min=max_calls_per_min,
            suppressed_exceptions=suppressed_exceptions,
            use_n_param=use_n_param,
        )

    async def evaluate(
        self, count: int = 25, metrics: MetricTypes = None, return_data: bool = False
    ) -> Dict[str, Dict[str, float]]:
        """
        Compute all the metrics based on the provided data.

        Parameters
        ----------
        count : int, default=25
            Specifies number of responses to generate for each prompt. The convention is to use 25
            generations per prompt in evaluating toxicity. See, for example DecodingTrust (https://arxiv.org/abs//2306.11698)
            or Gehman et al., 2020 (https://aclanthology.org/2020.findings-emnlp.301/).

        metrics : dict or list of str, optional
            Specifies which metrics to evaluate. If None, computes all supported metrics.

        return_data : bool, default=False
            Indicates whether to include response-level scores in results dictionary returned by this method.

        Returns
        -------
        dict
            A dictionary containing values of toxicity, stereotype, and counterfactual metrics and, optionally,
            response-level scores.
        """
        if metrics is not None:
            self.metrics = self._validate_metrics(metrics)

        print("\033[1mStep 1: Fairness Through Unawareness Check\033[0m")
        print("------------------------------------------")
        # 1. Check for Fairness Through Unawareness FTU
        # Parse prompts for protected attribute words
        protected_words = {"race": 0, "gender": 0}
        total_protected_words = 0
        for attribute in protected_words.keys():
            col = self.cf_generator_object.parse_texts(
                texts=self.prompts, attribute=attribute
            )
            protected_words[attribute] = sum(
                [1 if len(col_item) > 0 else 0 for col_item in col]
            )
            total_protected_words += protected_words[attribute]
            print(
                f"""Number of prompts containing {attribute} words: {protected_words[attribute]}"""
            )

        if total_protected_words > 0:
            print(
                "Fairness through unawareness is not satisfied. Toxicity, stereotype, and counterfactual fairness assessments will be conducted."
            )
            print("\n\033[1mStep 2: Generate Counterfactual Dataset\033[0m")
            print("---------------------------------------")
            # 2. Generate CF responses for race (if race FTU not satisfied) and gender (if gender FTU not satisfied)
            if (self.counterfactual_responses is None) and (
                "counterfactual" in self.metrics
            ):
                self.counterfactual_responses = {}
                self.counterfactual_response_metadata = {}
                for attribute in protected_words.keys():
                    if protected_words[attribute] > 0:
                        try:
                            self.counterfactual_responses[
                                attribute
                            ] = await self.cf_generator_object.generate_responses(
                                count=count, prompts=self.prompts, attribute=attribute
                            )
                        except AssertionError as e:
                            # Handle case where prompts don't contain the specific attribute words
                            print(
                                f"Warning: Could not generate counterfactual responses for {attribute}: {e}"
                            )
                            # Remove this attribute from protected_words to prevent KeyError later
                            protected_words[attribute] = 0
        else:
            print(
                "Fairness through unawareness is satisfied. Toxicity and stereotype assessments will be conducted."
            )
            print("\n\033[1m(Skipping) Step 2: Generate Counterfactual Dataset\033[0m")
            print("--------------------------------------------------")

        # 3. Generate responses for toxicity and stereotype evaluation (if responses not provided)
        if self.responses is None:
            print("\n\033[1mStep 3: Generating Model Responses\033[0m")
            print("----------------------------------")
            dataset = await self.generator_object.generate_responses(
                prompts=self.prompts,
                count=count,
            )
            self.prompts = dataset["data"]["prompt"]
            self.responses = dataset["data"]["response"]
        else:
            print("\n\033[1m(Skipping) Step 3: Generating Model Responses\033[0m")
            print("---------------------------------------------")

        # 4. Calculate toxicity metrics
        print("\n\033[1mStep 4: Evaluate Toxicity Metrics\033[0m")
        print("---------------------------------")
        toxicity_object = ToxicityMetrics(device=self.toxicity_device)
        toxicity_results = toxicity_object.evaluate(
            prompts=list(self.prompts), responses=list(self.responses), return_data=True
        )
        self.results["metrics"]["Toxicity"] = toxicity_results["metrics"]

        del toxicity_results["data"]["response"], toxicity_results["data"]["prompt"]
        self.toxicity_scores = toxicity_results["data"]
        del toxicity_results

        # 5. Calculate stereotype metrics
        print("\n\033[1mStep 5: Evaluate Stereotype Metrics\033[0m")
        print("-----------------------------------")
        attributes = [
            attribute
            for attribute in protected_words.keys()
            if protected_words[attribute] > 0
        ]
        stereotype_object = StereotypeMetrics(
            _classifier_model=self._stereotype_classifier_model
        )
        stereotype_results = stereotype_object.evaluate(
            prompts=list(self.prompts),
            responses=list(self.responses),
            return_data=True,
            categories=attributes,
        )
        self.results["metrics"]["Stereotype"] = stereotype_results["metrics"]

        del stereotype_results["data"]["response"], stereotype_results["data"]["prompt"]
        self.stereotype_scores = stereotype_results["data"]
        del stereotype_results

        # 6. Calculate CF metrics (if FTU not satisfied and counterfactual metrics requested)
        if total_protected_words > 0 and "counterfactual" in self.metrics:
            print("\n\033[1mStep 6: Evaluate Counterfactual Metrics\033[0m")
            print("---------------------------------------")
            print("Evaluating metrics...")
            self.results["metrics"]["Counterfactual"] = {}
            self.counterfactual_data = {}
            counterfactual_object = CounterfactualMetrics(
                neutralize_tokens=self.neutralize_tokens,
                transformer=self.counterfactual_transformer,
                sentiment_classifier=self.counterfactual_sentiment_classifier,
            )
            for attribute in Protected_Attributes.keys():
                if (
                    protected_words[attribute] > 0
                    and attribute in self.counterfactual_responses
                ):
                    for group1, group2 in combinations(
                        Protected_Attributes[attribute], 2
                    ):
                        group1_response = self.counterfactual_responses[attribute][
                            "data"
                        ][group1 + "_response"]
                        group2_response = self.counterfactual_responses[attribute][
                            "data"
                        ][group2 + "_response"]
                        successful_response_index = self._get_success_indices(
                            group1_response=group1_response,
                            group2_response=group2_response,
                        )
                        cf_group_results = counterfactual_object.evaluate(
                            texts1=[
                                group1_response[i] for i in successful_response_index
                            ],
                            texts2=[
                                group2_response[i] for i in successful_response_index
                            ],
                            attribute=attribute,
                            return_data=True,
                        )
                        self.results["metrics"]["Counterfactual"][
                            f"{group1}-{group2}"
                        ] = cf_group_results["metrics"]
                        self.counterfactual_data[f"{group1}-{group2}"] = (
                            cf_group_results["data"]
                        )
        else:
            print("\n\033[1m(Skipping) Step 6: Evaluate Counterfactual Metrics\033[0m")
            print("--------------------------------------------------")
            # Initialize empty counterfactual_data when metrics are skipped
            self.counterfactual_data = {}

        if return_data:
            self.results["data"]["Toxicity"] = self.toxicity_data
            self.results["data"]["Stereotype"] = self.stereotype_data
            self.results["data"]["Counterfactual"] = self.counterfactual_data

        return self.results

    @property
    def toxicity_data(self):
        self.toxicity_scores["prompt"] = self.prompts
        self.toxicity_scores["response"] = self.responses
        return self.toxicity_scores

    @property
    def stereotype_data(self):
        self.stereotype_scores["prompt"] = self.prompts
        self.stereotype_scores["response"] = self.responses
        return self.stereotype_scores

    def print_results(self) -> None:
        """
        Print the evaluate metrics values in the desired format.
        """
        result_list = self._create_result_list()
        print("".join(result_list))

    def export_results(self, file_name: str = "results.txt") -> None:
        """
        Export the evaluated metrics values in a text file.

        Parameters
        ----------
        file_name : str, Default = "results.txt"
            Name of the .txt file.
        """
        result_list = self._create_result_list(bold_headings=False)

        with open(file_name, "w+") as file:
            # Writing data to a file
            file.writelines(result_list)

    def _get_success_indices(
        self, group1_response: List[str], group2_response: List[str]
    ) -> List[any]:
        se = self.cf_generator_object.suppressed_exceptions
        if isinstance(se, Dict):
            failure_messages = set(self.suppressed_exceptions.values())
            failure_messages.add(FAILURE_MESSAGE)
            successful_response_index = [
                i
                for i in range(len(group1_response))
                if group1_response[i] not in failure_messages
                and group2_response[i] not in failure_messages
            ]
        else:
            successful_response_index = [
                i
                for i in range(len(group1_response))
                if group1_response[i] != FAILURE_MESSAGE
                and group2_response[i] != FAILURE_MESSAGE
            ]
        return successful_response_index

    def _create_result_list(self, bold_headings=True) -> List[str]:
        """Helper function for `print_results` method."""
        result_list = []
        start_heading, end_heading = "", ""
        if bold_headings:
            start_heading, end_heading = "\033[1m", "\033[0m"
        result_list.append(
            start_heading + "1. Toxicity Assessment" + end_heading + " \n"
        )
        for key in self.results["metrics"]["Toxicity"]:
            result_list.append(
                "- {:<40} {:1.4f} \n".format(
                    key, self.results["metrics"]["Toxicity"][key]
                )
            )

        result_list.append(
            start_heading + "2. Stereotype Assessment" + end_heading + " \n"
        )
        for key in self.results["metrics"]["Stereotype"]:
            tmp = "- {:<40} {:1.4f} \n"
            if self.results["metrics"]["Stereotype"][key] is None:
                tmp = "- {:<40} {} \n"
            result_list.append(
                tmp.format(key, self.results["metrics"]["Stereotype"][key])
            )

        if "Counterfactual" in self.results["metrics"]:
            result_list.append(
                start_heading + "3. Counterfactual Assessment" + end_heading + " \n"
            )
            tmp = ["{:<25}".format(" ")]
            for key in self.results["metrics"]["Counterfactual"]:
                tmp.append("{:<15}".format(key))
            tmp.append(" \n")
            result_list.append("".join(tmp))

            for metric_name in list(self.results["metrics"]["Counterfactual"].values())[
                0
            ]:
                tmp = ["- ", "{:<25}".format(metric_name)]
                for key in self.results["metrics"]["Counterfactual"]:
                    tmp.append(
                        "{:<15}".format(
                            "{:1.4f}".format(
                                self.results["metrics"]["Counterfactual"][key][
                                    metric_name
                                ]
                            )
                        )
                    )
                tmp.append(" \n")
                result_list.append("".join(tmp))
        return result_list

    def _validate_metrics(self, metrics: List[str]) -> None:
        """Validate that specified metrics are supported."""
        if metrics is None:
            metrics = DefaultMetrics
        elif isinstance(metrics, list):
            tmp = dict()
            for metric in metrics:
                if metric in DefaultMetrics:
                    tmp[metric] = DefaultMetrics[metric]
                else:
                    raise RuntimeError(
                        "If `metrics` is a list, it should be a subset of following list ['counterfactual', 'stereotype', 'toxicity']"
                    )
            metrics = tmp
        elif isinstance(metrics, dict):
            for key in metrics.keys():
                if key not in DefaultMetrics.keys():
                    raise KeyError("{} not found".format(key))
                self._check_list(
                    metrics[key], DefaultMetrics[key], "metrics['" + key + "']"
                )
        else:
            raise TypeError(
                "Attribute `metrics` should be a list of strings or a dictionary of list of strings"
            )
        return metrics

    @staticmethod
    def _validate_list_type(input_variable: List[str]) -> List[str]:
        """Validate inputs."""
        if isinstance(input_variable, list):
            if len(input_variable) == 0 or not isinstance(input_variable[0], str):
                raise RuntimeError(
                    "List {} should contain strings and can't be empty.".format(
                        input_variable
                    )
                )
        return input_variable

    @staticmethod
    def _check_list(list1: List[str], list2: List[str], error_tag: Any) -> None:
        """
        Check if list1 is a subset of list 2.
        """
        for list1_i in list1:
            if not isinstance(list1_i, str):
                raise TypeError(
                    "Type of list '{}' should be a string.".format(error_tag)
                )
            elif list1_i not in list2:
                raise RuntimeError(
                    "Provided '{}' metric is not an in-built langfair metric".format(
                        error_tag
                    )
                )



# Copyright 2024 CVS Health and/or one of its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from langfair.metrics.recommendation.metrics.baseclass.metrics import Metric


class PRAG(Metric):
    def __init__(self) -> None:
        """
        This class computes Pairwise Ranking Accuracy Gap (PRAG).
        """
        self.name = "PRAG"

    def evaluate(self, list1: List[str], list2: List[str]) -> float:
        """
        This method computes PRAG score between two lists.

        Parameters
        ----------
        list1 : list of strings
            A list of recommendation from an LLM model.

        list2 : list of strings
            Another list of recommendation from an LLM model.

        Returns
        -------
        float
            PRAG metric value for the two provided lists of recommendations (float)
        """
        K = len(list1)
        if not list1 or not list2:
            return 0

        if len(list1) == 1:
            return int(list1 == list2)

        matches = 0
        for i, item1 in enumerate(list1):
            for j, item2 in enumerate(list1[i + 1 :], start=i + 1):
                id1, id2 = (
                    list2.index(item) if item in list2 else -1
                    for item in (item1, item2)
                )
                if id1 != -1 and (id1 < id2 or id2 == -1):
                    matches += 1

        return matches / (K * (K + 1))



# Copyright 2024 CVS Health and/or one of its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import statistics
from abc import ABC, abstractmethod
from typing import Any, Dict, List


class Metric(ABC):
    """
    Abstract base class of fractaion, expected maximum, and probability metrics. Serves as a
    template for creating new metric functions.
    """

    @abstractmethod
    def metric_function(self, scores: List[float], toxic_threshold: float) -> float:
        """
        Abstract method that needs to be implemented by the user when creating a new metric function.
        """
        pass

    def evaluate(
        self, data: Dict[str, Any], threshold: float, score_column: str = "score"
    ) -> float:
        """
        This method compute metric function for unique input prompts and return the mean value over all
        unique prompts.

        Parameters
        ----------
        data: dict
            A dictionary containing toxicity/stereotype score.

        threshold : float
            Specifies the threshold to use for toxicity/stereotype score classification.

        score_column : str, default='score'
            Name of the dictionary key that contains score.
        """
        results = []
        for prompt in set(data["prompt"]):
            score = [
                data[score_column][i]
                for i in range(len(data[score_column]))
                if data["prompt"][i] == prompt
            ]
            results.append(self.metric_function(score, threshold))
        return statistics.mean(results)



# Copyright 2024 CVS Health and/or one of its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import numpy as np
from numpy.typing import ArrayLike

from langfair.metrics.classification.metrics.baseclass.metrics import Metric


class PredictedPrevalenceRateParity(Metric):
    def __init__(self) -> None:
        """
        This class computes predicted prevalence rate parity. The user may specify whether to compute this
        metric as a difference or a ratio. For more information on these metrics, see Feldman et al. (2015) :footcite:`feldman2015certifyingremovingdisparateimpact`.
        """
        self.name = "PredictedPrevalenceRateParity"

    def evaluate(
        self,
        groups: ArrayLike,
        y_pred: ArrayLike,
        y_true: Optional[ArrayLike] = None,
        ratio: bool = False,
    ) -> float:
        """
        This method computes disparity in predicted prevalence rates between two groups.

        Parameters
        ----------
        groups : Array-like
            Group indicators. Must contain exactly two unique values.

        y_pred : Array-like
            Binary model predictions. Positive and negative predictions must be 1 and 0, respectively.

        y_true : Array-like, default=None
            Binary labels (ground truth values). This argument is only a placeholder for this class.

        ratio : bool, default=False
            Indicates whether to compute the metric as a difference or a ratio

        Returns
        -------
        float
            Value of predicted prevalence rate parity

        References
        ----------
        .. footbibliography::
        """
        unique_preds = np.unique(y_pred)
        assert np.array_equal(unique_preds, [0, 1]), (
            "y_pred must contain exactly two unique values: 0 and 1"
        )

        unique_groups = np.unique(groups)
        assert len(unique_groups) == 2, "groups must contain exactly two unique values"

        ppr1 = np.mean(y_pred[groups == unique_groups[0]])
        ppr2 = np.mean(y_pred[groups == unique_groups[1]])

        return ppr1 / ppr2 if ratio else abs(ppr1 - ppr2)
